package lcsb.mapviewer.web.config;

import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.security.authentication.AuthenticationProvider;
import org.springframework.security.config.annotation.authentication.builders.AuthenticationManagerBuilder;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.builders.WebSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.config.annotation.web.configuration.WebSecurityConfigurerAdapter;
import org.springframework.security.config.http.SessionCreationPolicy;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.session.SessionInformation;
import org.springframework.security.core.session.SessionRegistry;
import org.springframework.security.web.authentication.AuthenticationFailureHandler;
import org.springframework.security.web.authentication.AuthenticationSuccessHandler;
import org.springframework.security.web.authentication.Http403ForbiddenEntryPoint;
import org.springframework.security.web.authentication.logout.LogoutHandler;
import org.springframework.security.web.authentication.logout.LogoutSuccessHandler;
import org.springframework.security.web.authentication.www.BasicAuthenticationFilter;

import lcsb.mapviewer.api.SpringRestApiConfig;
import lcsb.mapviewer.services.SpringServiceConfig;
import lcsb.mapviewer.web.bean.utils.CORSFilter;
import lcsb.mapviewer.web.bean.utils.CacheFilter;
import lcsb.mapviewer.web.bean.utils.XFrameFilter;

@Configuration
@ComponentScan(basePackages = { "lcsb.mapviewer.web.config" })
@Import({ SpringRestApiConfig.class, SpringServiceConfig.class })
@EnableWebSecurity
public class SpringSecurityConfig extends WebSecurityConfigurerAdapter {

  private static final Logger logger = LogManager.getLogger();

  private AuthenticationSuccessHandler successHandler;
  private AuthenticationFailureHandler failureHandler;
  private LogoutSuccessHandler logoutSuccessHandler;
  private List<AuthenticationProvider> authenticationProviders;

  private SessionRegistry sessionRegistry;

  @Autowired
  SpringSecurityConfig(final AuthenticationSuccessHandler successHandler,
      final AuthenticationFailureHandler failureHandler,
      final LogoutSuccessHandler logoutSuccessHandler,
      final List<AuthenticationProvider> authenticationProviders,
      final SessionRegistry sessionRegistry) {
    super(false);
    this.successHandler = successHandler;
    this.failureHandler = failureHandler;
    this.logoutSuccessHandler = logoutSuccessHandler;
    this.authenticationProviders = authenticationProviders;
    this.sessionRegistry = sessionRegistry;
  }

  @Override
  protected void configure(final AuthenticationManagerBuilder auth) {
    authenticationProviders.forEach(auth::authenticationProvider);
  }

  @Override
  public void configure(final WebSecurity web) {
    web.ignoring()
        .antMatchers("/")
        .antMatchers("*.xhtml")
        .antMatchers("/resources/**")
        .antMatchers("/javax.faces.resource/**");
  }

  @Override
  protected void configure(final HttpSecurity http) throws Exception {
    http
        .sessionManagement()
        .maximumSessions(-1).sessionRegistry(sessionRegistry)
        .and()
        .sessionCreationPolicy(SessionCreationPolicy.IF_REQUIRED)
        .and()
        .anonymous().principal(lcsb.mapviewer.common.Configuration.ANONYMOUS_LOGIN)
        .and()
        .exceptionHandling()
        .authenticationEntryPoint(new Http403ForbiddenEntryPoint())
        .and()
        .formLogin()
        .usernameParameter("login")
        .passwordParameter("password")
        .loginProcessingUrl("/api/doLogin")
        .successHandler(successHandler)
        .failureHandler(failureHandler)
        .and()
        .logout()
        .logoutUrl("/api/doLogout").permitAll()
        .logoutSuccessHandler(logoutSuccessHandler)
        .addLogoutHandler(new LogoutHandler() {
          @Override
          public void logout(final HttpServletRequest request, final HttpServletResponse response, final Authentication authentication) {
            try {
              // for some reason session registry is not informed about logout
              SessionInformation sessionInformation = sessionRegistry.getSessionInformation(request.getSession().getId());
              if (sessionInformation != null) {
                sessionInformation.expireNow();
              }
            } catch (Exception e) {
              logger.error(e, e);
            }
          }
        })
        .deleteCookies(lcsb.mapviewer.common.Configuration.AUTH_TOKEN)
        .invalidateHttpSession(true)
        .and()
        .authorizeRequests()
        .and()
        .headers()
        .frameOptions().disable() // is managed by XFrameFilter
        .and()
        .cors().disable() // is managed by CORSFilter
        .csrf().disable()
        .addFilterAfter(new XFrameFilter(), BasicAuthenticationFilter.class)
        .addFilterAfter(new CacheFilter(), BasicAuthenticationFilter.class)
        .addFilterAfter(new CORSFilter(), BasicAuthenticationFilter.class);
  }

}