package lcsb.mapviewer.web;

import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.documentationConfiguration;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.delete;
import static org.springframework.restdocs.request.RequestDocumentation.parameterWithName;
import static org.springframework.restdocs.request.RequestDocumentation.pathParameters;
import static org.springframework.restdocs.request.RequestDocumentation.requestParameters;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.awt.geom.Point2D;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.servlet.Filter;

import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.core.LogEvent;
import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.restdocs.JUnitRestDocumentation;
import org.springframework.restdocs.cli.CliDocumentation;
import org.springframework.restdocs.http.HttpDocumentation;
import org.springframework.restdocs.operation.preprocess.Preprocessors;
import org.springframework.restdocs.payload.PayloadDocumentation;
import org.springframework.restdocs.request.PathParametersSnippet;
import org.springframework.restdocs.request.RequestParametersSnippet;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.web.WebAppConfiguration;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.RequestBuilder;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import org.springframework.web.context.WebApplicationContext;

import com.fasterxml.jackson.databind.ObjectMapper;

import lcsb.mapviewer.annotation.data.serializer.ChemicalSerializer;
import lcsb.mapviewer.common.MinervaLoggerAppender;
import lcsb.mapviewer.common.TextFileUtils;
import lcsb.mapviewer.common.tests.TestUtils;
import lcsb.mapviewer.common.tests.UnitTestFailedWatcher;
import lcsb.mapviewer.converter.ColorSchemaReader;
import lcsb.mapviewer.converter.zip.ZipEntryFileFactory;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.ProjectLogEntry;
import lcsb.mapviewer.model.ProjectLogEntryType;
import lcsb.mapviewer.model.ProjectStatus;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.graphics.HorizontalAlign;
import lcsb.mapviewer.model.graphics.PolylineData;
import lcsb.mapviewer.model.graphics.VerticalAlign;
import lcsb.mapviewer.model.map.Comment;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.OverviewImage;
import lcsb.mapviewer.model.map.OverviewImageLink;
import lcsb.mapviewer.model.map.OverviewModelLink;
import lcsb.mapviewer.model.map.OverviewSearchLink;
import lcsb.mapviewer.model.map.compartment.Compartment;
import lcsb.mapviewer.model.map.compartment.SquareCompartment;
import lcsb.mapviewer.model.map.kinetics.SbmlFunction;
import lcsb.mapviewer.model.map.kinetics.SbmlKinetics;
import lcsb.mapviewer.model.map.kinetics.SbmlParameter;
import lcsb.mapviewer.model.map.kinetics.SbmlUnit;
import lcsb.mapviewer.model.map.layout.ProjectBackground;
import lcsb.mapviewer.model.map.layout.ProjectBackgroundImageLayer;
import lcsb.mapviewer.model.map.layout.ProjectBackgroundStatus;
import lcsb.mapviewer.model.map.layout.graphics.Glyph;
import lcsb.mapviewer.model.map.model.ElementSubmodelConnection;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.ModelSubmodelConnection;
import lcsb.mapviewer.model.map.model.SubmodelType;
import lcsb.mapviewer.model.map.reaction.NodeOperator;
import lcsb.mapviewer.model.map.reaction.Product;
import lcsb.mapviewer.model.map.reaction.Reactant;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.reaction.SplitOperator;
import lcsb.mapviewer.model.map.reaction.type.TransportReaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.GenericProtein;
import lcsb.mapviewer.model.map.species.SimpleMolecule;
import lcsb.mapviewer.model.overlay.DataOverlay;
import lcsb.mapviewer.model.overlay.DataOverlayType;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.persist.DbUtils;
import lcsb.mapviewer.persist.dao.ProjectDao;
import lcsb.mapviewer.services.interfaces.ICommentService;
import lcsb.mapviewer.services.interfaces.IDataOverlayService;
import lcsb.mapviewer.services.interfaces.IFileService;
import lcsb.mapviewer.services.interfaces.IProjectService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.web.config.SpringWebConfig;
import lcsb.mapviewer.web.utils.CommandFormatterWithReplacingPostFilenameHeader;
import lcsb.mapviewer.web.utils.CustomCurlRequestSnippet;

@WebAppConfiguration
@ContextConfiguration(classes = SpringWebConfig.class)
public abstract class ControllerIntegrationTest extends TestUtils {

  protected static String INVALID_LONG_NAME = "aaaaaaaaxvncnbvscbnmcnbmccbnsbnsdsnbmdsvb"
      + "nmsdvnbmsdbmnbndvmsbnmsvdnbmnmbdsvnbmdsvxncbmbnmscbnzdnbnabnsbnamsdbmnsadbmnasd"
      + "bnmnbmsadbnmasdnbasdbnmsadnbnbmsadbnmadsnbmadsnbnbsadnbmadsbnndsabnbmdasbnmdsaj"
      + "qwrhgjrwhjghgjwerghjwreghwewnjnnbbbnbnbmbnbnzcmnnbmzcnmbcsbnmcsnbcnbzmnbczxnbmc"
      + "zxnbmcxznbcnxbmznbmxzcnbzcxnnbcxznbmzcnbczxnbmnbzcxnbmcznnczbnbzcnbmzcbnmbncznb"
      + "cznbcz";

  protected static Logger logger = LogManager.getLogger();

  @Rule
  public JUnitRestDocumentation junitRestDocumentation = new JUnitRestDocumentation();

  @Rule
  public UnitTestFailedWatcher unitTestFailedWatcher = new UnitTestFailedWatcher();
  protected MockMvc mockMvc;
  @Autowired
  private WebApplicationContext context;
  @Autowired
  private Filter springSecurityFilterChain;

  @Autowired
  private IUserService userService;

  @Autowired
  private IProjectService projectService;

  @Autowired
  private ProjectDao projectDao;

  @Autowired
  private ICommentService commentService;

  @Autowired
  private PasswordEncoder passwordEncoder;

  @Autowired
  private IFileService fileService;

  @Autowired
  private IDataOverlayService dataOverlayService;

  @Autowired
  protected DbUtils dbUtils;
  private MinervaLoggerAppender appender;

  private ExecutorService executorService;

  @Autowired
  protected ObjectMapper objectMapper;

  @PostConstruct
  public void construct() {
    mockMvc = MockMvcBuilders.webAppContextSetup(context)
        .addFilter(springSecurityFilterChain)
        .build();

    mockMvc = MockMvcBuilders.webAppContextSetup(context)
        .addFilter(springSecurityFilterChain)
        .apply(documentationConfiguration(this.junitRestDocumentation)
            .operationPreprocessors()
            .withResponseDefaults(Preprocessors.prettyPrint())
            .and()
            .uris().withHost("minerva-service.lcsb.uni.lu/minerva").withPort(443).withScheme("https")
            .and().snippets().withDefaults(
                new CustomCurlRequestSnippet(null, new CommandFormatterWithReplacingPostFilenameHeader(" \\%n    ")),
                CliDocumentation.httpieRequest(),
                HttpDocumentation.httpRequest(),
                HttpDocumentation.httpResponse(),
                PayloadDocumentation.requestBody(),
                PayloadDocumentation.responseBody()))
        .build();

    executorService = Executors.newFixedThreadPool(1);
  }

  @PreDestroy
  public final void __tearDown() throws Exception {
    executorService.shutdown();
  }

  /**
   * This method can be used to work around the fact that MockMvc cannot
   * retrieve cookies from Spring Security. The Reason for that is that MockMvc
   * calls the controller directly and (partially?) bypasses Spring.
   * <p>
   * This method creates a mocked session that can be used in a request as
   * opposed to the token.
   * </p>
   * <p>
   * FIXME: Find a better solution, that does not violate the spirit of
   * integration tests.
   * </p>
   */
  protected MockHttpSession createSession(final String login, final String password) throws Exception {
    RequestBuilder request = post("/api/doLogin")
        .param("login", login)
        .param("password", password);
    return (MockHttpSession) mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn()
        .getRequest()
        .getSession();
  }

  @Before
  public final void _setUp() throws Exception {
    MinervaLoggerAppender.unregisterLogEventStorage(appender);
    appender = MinervaLoggerAppender.createAppender();
  }

  @After
  public final void _tearDown() throws Exception {
    MinervaLoggerAppender.unregisterLogEventStorage(appender);
  }

  protected List<LogEvent> getWarningLogs() {
    return appender.getWarnings();
  }

  protected List<LogEvent> getErrorLogs() {
    return appender.getErrors();
  }

  protected List<LogEvent> getInfoLogs() {
    return appender.getInfos();
  }

  protected List<LogEvent> getDebugLogs() {
    return appender.getDebugs();
  }

  protected User createUser(final String login, final String password) {
    User user = new User();
    user.setLogin(login);
    user.setCryptedPassword(passwordEncoder.encode(password));
    userService.addUser(user);
    return user;
  }

  protected User createUser(final String login, final String password, final Project project) {
    User user = createUser(login, password);
    userService.grantUserPrivilege(user, PrivilegeType.READ_PROJECT, project.getProjectId());
    return user;
  }

  protected User createAdmin(final String login, final String password) {
    User user = createUser(login, password);
    userService.grantUserPrivilege(user, PrivilegeType.IS_ADMIN);
    return user;
  }

  protected User createCurator(final String login, final String password) {
    User user = createUser(login, password);
    userService.grantUserPrivilege(user, PrivilegeType.IS_CURATOR);
    return user;
  }

  protected User createCurator(final String login, final String password, final Project project) {
    User user = createCurator(login, password);
    userService.grantUserPrivilege(user, PrivilegeType.READ_PROJECT, project.getProjectId());
    userService.grantUserPrivilege(user, PrivilegeType.WRITE_PROJECT, project.getProjectId());
    return user;
  }

  protected Project createProjectWithGlyph(final String projectId) throws IOException {
    User admin = userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN);
    Project project = new Project(projectId);
    project.setOwner(admin);
    UploadedFileEntry file = createFile(Files.readAllBytes(Paths.get("./src/test/resources/empty.png")), admin);
    ModelData map = new ModelData();
    map.setTileSize(256);
    map.setWidth(100);
    map.setHeight(100);

    Element element = createProtein();
    Glyph glyph = new Glyph();
    glyph.setFile(file);
    element.setGlyph(glyph);
    map.addElement(element);

    project.addModel(map);
    project.addGlyph(glyph);
    projectService.add(project);
    return project;
  }

  protected Project createAndPersistProject(final String projectId) {
    Project project = new Project(projectId);
    project.setOwner(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    project.setVersion("1.0");
    project.setDisease(new MiriamData(MiriamType.MESH_2012, "D010300"));
    project.setOrganism(new MiriamData(MiriamType.TAXONOMY, "9606"));
    project.setDirectory("289b78b436176091ad900020c933c544");
    project.setNotifyEmail("minerva@uni.lu");
    OverviewImage image = new OverviewImage();
    image.setFilename("image.png");
    image.setHeight(100);
    image.setWidth(200);
    project.addOverviewImage(image);
    project.setName("Test Disease");
    ProjectLogEntry entry = new ProjectLogEntry();
    entry.setMapName("map_name");
    entry.setSeverity("WARNING");
    entry.setType(ProjectLogEntryType.OTHER);
    entry.setContent("This is warning message");
    project.addLogEntry(entry);
    try {
      project.setInputData(createFile(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")),
          userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN)));
    } catch (final IOException e) {
      e.printStackTrace();
    }

    ModelData map = new ModelData();
    OverviewModelLink link1 = new OverviewModelLink();
    link1.setLinkedModel(map);
    link1.setPolygon("0,0 10,10 10,0");
    link1.setxCoord(10);
    link1.setyCoord(20);
    link1.setZoomLevel(2);
    image.addLink(link1);

    OverviewImageLink link2 = new OverviewImageLink();
    link2.setLinkedOverviewImage(image);
    link2.setPolygon("0,0 10,10 10,0");
    image.addLink(link2);

    OverviewSearchLink link3 = new OverviewSearchLink();
    link3.setQuery("a1");
    link3.setPolygon("0,0 10,10 10,0");
    image.addLink(link3);

    map.setName("map_name");
    map.setTileSize(256);
    map.setWidth(100);
    map.setHeight(100);
    SbmlFunction sbmlFunction = new SbmlFunction("f1");
    sbmlFunction.setName("name");
    sbmlFunction.setDefinition("<math xmlns=\"http://www.w3.org/1998/Math/MathML\"><lambda>\n"
        + "<bvar>\n"
        + "<ci> x </ci>\n"
        + "</bvar>\n"
        + "<bvar>\n"
        + "<ci> y </ci>\n"
        + "</bvar>\n"
        + "<apply>\n"
        + "<plus/>\n"
        + "<ci> x </ci>\n"
        + "<ci> y </ci>\n"
        + "</apply>\n"
        + "</lambda></math>");
    map.addFunction(sbmlFunction);

    SbmlParameter sbmlParameter = new SbmlParameter("param1");
    sbmlParameter.setName("ppp");
    sbmlParameter.setValue(1.2);
    map.addParameter(sbmlParameter);

    SbmlUnit sbmlUnit = new SbmlUnit("u1");
    sbmlUnit.setName("uuu");
    map.addUnit(sbmlUnit);

    Compartment comp = createCompartment();
    map.addElement(comp);

    Reaction reaction = new TransportReaction("path_0_re5933");
    reaction.setZ(2);
    reaction.setLine(new PolylineData(new Point2D.Double(0, 0), new Point2D.Double(10, 0)));
    reaction.addMiriamData(new MiriamData(MiriamType.PUBMED, "28725475"));
    SbmlKinetics kinetics = new SbmlKinetics();
    kinetics.setDefinition("<math xmlns=\"http://www.w3.org/1998/Math/MathML\"><lambda>\n"
        + "<bvar>\n"
        + "<ci> x </ci>\n"
        + "</bvar>\n"
        + "<bvar>\n"
        + "<ci> y </ci>\n"
        + "</bvar>\n"
        + "<apply>\n"
        + "<plus/>\n"
        + "<ci> x </ci>\n"
        + "<ci> y </ci>\n"
        + "</apply>\n"
        + "</lambda></math>");
    reaction.setKinetics(kinetics);
    map.addReaction(reaction);

    Element element = createProtein();
    map.addElement(element);
    comp.addElement(element);

    Element element2 = createSimpleMolecule();
    map.addElement(element2);
    comp.addElement(element2);

    Element element3 = createSimpleMolecule();
    map.addElement(element3);

    Reactant reactant = new Reactant(element2);
    reactant.setLine(new PolylineData(new Point2D.Double(10, 0), new Point2D.Double(0, 0)));
    reaction.addReactant(reactant);

    Product product = new Product(element2);
    product.setLine(new PolylineData(new Point2D.Double(15, 0), new Point2D.Double(20, 0)));
    reaction.addProduct(product);

    Product product2 = new Product(element3);
    product2.setLine(new PolylineData(new Point2D.Double(15, 0), new Point2D.Double(30, 0)));
    reaction.addProduct(product2);

    NodeOperator operator = new SplitOperator();
    operator.addOutput(product);
    operator.addOutput(product2);
    operator.setLine(new PolylineData(new Point2D.Double(10, 0), new Point2D.Double(15, 0)));
    reaction.addNode(operator);

    ModelData submap = new ModelData();
    submap.setTileSize(256);
    submap.setWidth(100);
    submap.setHeight(100);
    submap.addElement(createProtein());

    ModelSubmodelConnection submodelConnection = new ModelSubmodelConnection(submap, SubmodelType.UNKNOWN);
    map.addSubmodel(submodelConnection);
    element.setSubmodel(new ElementSubmodelConnection(submap, SubmodelType.UNKNOWN));

    project.addModel(map);
    project.addModel(submap);

    ProjectBackground background = new ProjectBackground("Normal");
    int id = 0;
    for (final ModelData model : project.getModels()) {
      ProjectBackgroundImageLayer layer = new ProjectBackgroundImageLayer(model, "dir_" + (id++));
      background.addProjectBackgroundImageLayer(layer);
    }
    background.setDefaultOverlay(true);
    background.setOrderIndex(0);
    background.setProgress(100);
    background.setStatus(ProjectBackgroundStatus.OK);
    background.setCreator(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    project.addProjectBackground(background);

    projectService.add(project);
    return project;
  }

  private Element createSimpleMolecule() {
    Element molecule = new SimpleMolecule("s" + (counter++));
    molecule.setName("water");
    assignCoordinates(10, 20, 100, 20, molecule);
    return molecule;
  }

  private int counter = 1;

  private int z = 10;

  private GenericProtein createProtein() {
    GenericProtein element = new GenericProtein("p" + (counter++));
    element.setName("GSTA4");
    assignCoordinates(10, 20, 100, 20, element);
    return element;
  }

  private Compartment createCompartment() {
    SquareCompartment element = new SquareCompartment("p" + (counter++));
    element.setName("Comp");
    assignCoordinates(5, 5, 1000, 1000, element);
    return element;
  }

  private void assignCoordinates(final double x, final double y, final double width, final double height, final Element protein) {
    protein.setX(x);
    protein.setY(y);
    protein.setZ(z++);
    protein.setWidth(width);
    protein.setHeight(height);
    protein.setNameX(x);
    protein.setNameY(y);
    protein.setNameWidth(width);
    protein.setNameHeight(height);
    protein.setNameVerticalAlign(VerticalAlign.MIDDLE);
    protein.setNameHorizontalAlign(HorizontalAlign.CENTER);
  }

  protected Project createEmptyProject(final String projectId) {
    Project project = new Project(projectId);
    project.setOwner(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    project.setDirectory("289b78b436176091ad900020c933c544");
    project.setName("Test Disease");
    try {
      project.setInputData(createFile(Files.readAllBytes(Paths.get("./src/test/resources/generic.xml")),
          userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN)));
    } catch (final IOException e) {
      e.printStackTrace();
    }

    ModelData map = new ModelData();
    map.setName("map_name");
    map.setTileSize(256);
    map.setWidth(100);
    map.setHeight(100);
    Element element = createProtein();
    element.setName("PTI");
    map.addElement(element);

    project.addModel(map);
    projectDao.add(project);
    return project;
  }

  protected UploadedFileEntry createFile(final String content, final User user) {
    return createFile(content.getBytes(), user);
  }

  protected UploadedFileEntry createFile(final byte[] content, final User user) {
    UploadedFileEntry file = new UploadedFileEntry();
    file.setFileContent(content);
    file.setOriginalFileName("test_file");
    file.setLength(content.length);
    file.setOwner(user);
    fileService.add(file);
    return file;
  }

  protected void removeUser(final User user) throws Exception {
    if (user != null) {
      userService.deleteUser(user);
    }
  }

  protected void removeFile(final UploadedFileEntry fileEntry) throws Exception {
    fileService.delete(fileEntry);
  }

  protected void waitForProjectToFinishLoading(final String projectId) throws InterruptedException {
    Project project;
    do {
      project = projectService.getProjectByProjectId(projectId);
      if (project == null) {
        logger.warn("Project doesn't exist: " + projectId);
        break;
      }
      Thread.sleep(100);
    } while (project.getStatus() != ProjectStatus.DONE && project.getStatus() != ProjectStatus.FAIL);
  }

  public void removeProject(final Project project) throws Exception {
    if (project != null) {
      removeProject(project.getProjectId());
    }
  }

  public void removeProject(final String projectId) throws Exception {
    Project project = projectService.getProjectByProjectId(projectId);
    if (project != null) {
      RequestBuilder request = delete("/api/projects/{projectId}/", projectId)
          .session(createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD));

      mockMvc.perform(request).andExpect(status().is2xxSuccessful());
    }
    projectService.waitForJobQueueToEmpty();

  }

  protected DataOverlay createOverlay(final Project project, final User admin) throws Exception {
    return createOverlay(project, admin, "element_identifier\tvalue\n\t-1");
  }

  protected DataOverlay createOverlay(final Project project, final User admin, final String content) throws Exception {
    return createOverlay(project, admin, content, DataOverlayType.GENERIC);
  }

  protected DataOverlay createOverlay(final Project project, final User admin, final String content, final DataOverlayType type)
      throws Exception {
    UploadedFileEntry file = new UploadedFileEntry();
    file.setFileContent(content.getBytes());
    file.setLength(content.getBytes().length);
    file.setOriginalFileName("test.txt");

    ColorSchemaReader reader = new ColorSchemaReader();

    DataOverlay overlay = new DataOverlay();
    overlay.setColorSchemaType(type);
    overlay.setInputData(file);
    overlay.setProject(project);
    overlay.setName("test title");
    overlay.setDescription("test description");
    overlay.setCreator(admin);
    Map<String, String> headers = new HashMap<>();
    headers.put(TextFileUtils.COLUMN_COUNT_PARAM, "2");
    headers.put(ZipEntryFileFactory.LAYOUT_HEADER_PARAM_TYPE, type.toString());
    overlay.addEntries(reader.readColorSchema(new ByteArrayInputStream(content.getBytes()), headers));
    dataOverlayService.add(overlay);
    return overlay;
  }

  protected DataOverlay createOverlay(final String projectId, final User user) throws Exception {
    Project project = projectService.getProjectByProjectId(projectId);
    return createOverlay(project, user);
  }

  protected DataOverlay createOverlayWithColor(final Project project, final User admin) throws Exception {
    return createOverlay(project, admin, "element_identifier\tcolor\n\t#aa00cc");
  }

  protected Comment createAndPersistComment(final ModelData model, final User owner) throws Exception {
    Comment comment = createComment(model, owner);
    return commentService.add(comment);
  }

  protected ModelData getBuildInModel() throws Exception {
    return projectService.getProjectByProjectId(BUILT_IN_PROJECT, true).getTopModelData();
  }

  protected Comment createComment(final ModelData map) {
    return createComment(map, null);
  }

  protected Comment createComment(final ModelData map, final User owner) {
    Comment comment = new Comment();
    comment.setModel(map);
    comment.setUser(owner);
    comment.setCoordinates(new Point2D.Double(10, 20));
    return comment;
  }

  protected PathParametersSnippet getProjectPathParameters() {
    return pathParameters(parameterWithName("projectId").description("project identifier"));
  }

  protected PathParametersSnippet getMapPathParameters() {
    return getProjectPathParameters().and(parameterWithName("mapId").description("map identifier"));
  }

  protected PathParametersSnippet getOverlayPathParameters() {
    return getProjectPathParameters().and(parameterWithName("overlayId").description("overlay identifier"));
  }

  protected PathParametersSnippet getOverlayMapPathParameters() {
    return getMapPathParameters().and(parameterWithName("overlayId").description("overlay identifier"));
  }

  protected RequestParametersSnippet getChemicalFilter() {
    return requestParameters(
        parameterWithName("columns")
            .description("set of columns (all by default). Available options: "
                + StringUtils.join(ChemicalSerializer.availableColumns(), ", "))
            .optional(),
        parameterWithName("query")
            .description("name of chemical that we are searching for")
            .optional(),
        parameterWithName("target")
            .description("target element that we are searching for in format TYPE:ID")
            .optional());
  }
}
